/*****************************************************************************/
/* tcpip.c                                                                   */
/*                                                                           */
/* A simple program implementing a TCP/IP relay controller for interfacing   */
/* Webots with any development environment able to use TCP/IP, including     */
/* MathLab, Lisp, Java, C, C++, etc.                                         */
/*                                                                           */
/* The protocole used in this example is taken from the Khepera serial       */
/* communication protocole. Hence, if you already have developed an          */
/* application which uses this protocole (and send the data over the serial  */
/* port), you will just need to redirect the data to the TCP/IP socket of    */
/* this controller to make it work with Webots.                              */
/*                                                                           */
/* Currently supported Khepera protocole commands include:                   */
/*                                                                           */
/* B: read software version                                                  */
/* D: set speed                                                              */
/* G: set position counter                                                   */
/* H: read position                                                          */
/* L: change LED state                                                       */
/* N: read proximity sensors                                                 */
/* O: read ambient light sensors                                             */
/*                                                                           */
/* A sample client program, written in C is included in this directory.      */
/* See client.c for the source code                                          */
/* compile it with gcc client.c -o client                                    */
/*                                                                           */
/* Everything relies on standard POSIX TCP/IP sockets.                       */
/*****************************************************************************/

/*****************************************************************************/
/* Slighly modified to support IVR course, Edinburgh University.             */
/* Darren Smith, October 2003                                                */
/*****************************************************************************/

/*****************************************************************************/
/* Feel free to modify this sample program to suit your needs                */
/*****************************************************************************/

#include <stdio.h>
#include <string.h> 
#include <time.h>
#include <sys/types.h>
#include <device/robot.h>
#include <device/differential_wheels.h>
#include <device/distance_sensor.h>
#include <device/light_sensor.h>
#include <device/led.h>
#include <device/gps.h>
#ifdef WIN32
#include <winsock.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h> /* definition of struct sockaddr_in */
#include <netdb.h>      /* definition of gethostbyname */
#include <arpa/inet.h>  /* definition of inet_ntoa */
#include <unistd.h>     /* definition of close */
#endif

#define SOCKET_PORT 10021
#define NB_IR_SENSOR 8
#define TIMESTEP 250


static DeviceTag distance[NB_IR_SENSOR];
static DeviceTag ambient[NB_IR_SENSOR];
static DeviceTag led[2];
static DeviceTag ds[16];
static DeviceTag gps;

static int fd;
static fd_set rfds;

/*----------------------------------------------------------------------*/
static int accept_client(int server_fd) {
  int cfd;
  struct sockaddr_in client;
  int asize = sizeof(struct sockaddr_in);
  struct hostent *client_info;

  cfd = accept(server_fd,(struct sockaddr *)&client,&asize);
  if (cfd==-1) {
    printf("cannot accept client\n");
    return -1;
  }
  client_info = gethostbyname((char *)inet_ntoa(client.sin_addr));
  printf("Accepted connection from: %s \n",client_info->h_name);
  return cfd;
}

/*----------------------------------------------------------------------*/
static int create_socket_server(int port) {
  int sfd,rc;
  struct sockaddr_in address;

#ifdef WIN32
  /* initialize the socket api */
  WSADATA info;
  rc = WSAStartup(MAKEWORD(1,1),&info); /* Winsock 1.1 */
  if (rc!=0) {
    printf("cannot initialize Winsock\n");
    return -1;
  }
#endif
  /* create the socket */
  sfd = socket(AF_INET,SOCK_STREAM,0);
  if (sfd==-1) {
    printf("cannot create socket\n");
    return -1;
  }
  /* fill in socket address */
  memset(&address,0,sizeof(struct sockaddr_in));
  address.sin_family = AF_INET;
  address.sin_port = htons((unsigned short)port);
  address.sin_addr.s_addr = INADDR_ANY;
  /* bind to port */
  rc = bind(sfd,(struct sockaddr *)&address,sizeof(struct sockaddr));
  if (rc==-1) {
    printf("cannot bind port %d\n",port);
    return -1;
  }
  /* listen for connections */
  if (listen(sfd,1)==-1) {
    printf("cannot listen for connections\n");
    return -1;
  }
  printf("Waiting for a connection on port %d...\n",port);
  return accept_client(sfd);
}

/*----------------------------------------------------------------------*/
static void reset(void) {
  int i;
  char text[32];
  for(i=0;i<NB_IR_SENSOR;i++) {
    sprintf(text,"ds%d",i);
    distance[i] = robot_get_device(text);
    distance_sensor_enable(distance[i],TIMESTEP);
  }
  for(i=0;i<NB_IR_SENSOR;i++) {
    sprintf(text,"ls%d",i);
    ambient[i] = robot_get_device(text);
    light_sensor_enable(ambient[i],TIMESTEP);
  }
  led[0]=robot_get_device("led0");
  led[1]=robot_get_device("led1");
  /* Enable the wheel encoders measurements */
  differential_wheels_enable_encoders(TIMESTEP);
  printf("Khepera robot has been reset by Webots\n");
  fd = create_socket_server(SOCKET_PORT);
  FD_ZERO(&rfds);
  FD_SET(fd,&rfds);

// sensores del pioneer
  const char *robot_name = robot_get_name();
  char sensor_name[32];
  if (strcmp(robot_name,"pioneer2")==0) {
    for(i=0;i<4;i++) {
      sprintf(sensor_name,"FR%d",i+1);
      ds[i] = robot_get_device(sensor_name);
    }
    for(i=4;i<8;i++) {
      sprintf(sensor_name,"RR%d",i-3);
      ds[i] = robot_get_device(sensor_name);
    }
    for(i=8;i<12;i++) {
      sprintf(sensor_name,"FL%d",i-7);
      ds[i] = robot_get_device(sensor_name);
    }
    for(i=12;i<16;i++) {
      sprintf(sensor_name,"RL%d",i-11);
      ds[i] = robot_get_device(sensor_name);
    }
  }
  robot_console_printf("reset: robot name is %s\n",robot_name);
  for (i=0;i<16;i++) distance_sensor_enable(ds[i],64);
}

// gps
  gps = robot_get_device("gps");
  gps_enable(gps,64);

static int run(int ms) {
  int n;
  int ret;
  char buffer[256];
  short left_speed,right_speed;
  short led_number,led_action;
  static short led_value[2]={0,0}; /* initially off */
  struct timeval tv={0,0};
  int number;
  const float *matrix;
  float *euler;

  /* Set up the parameters used for the select statement */
  
  /* Do a robot timestep - i.e., run the webots simulator so that TIMESTEP */
  /* time is simulated. */
  FD_ZERO(&rfds);
  FD_SET(fd,&rfds);
  /* Watch TCPIP file descriptor to see when it has input. */
  /* No wait - polling as fast as possible */
  number = select(fd+1,&rfds,NULL,NULL,&tv);
  /* If there is no data at the socket, then redo loop */
  if (number == 0) return TIMESTEP;
  
  /* ...otherwise, there is data to read, so read & process. */
  n = recv(fd,buffer,256,0);
  if (n<0) printf("error reading from socket\n");
  buffer[n]='\0';
  printf("Received %d bytes: %s\n",n,buffer);


  if(buffer[0]=='A') {/* set the speed of the motors */
     differential_wheels_set_speed(1,1)
 } else if(buffer[0]=='R') {
     differential_wheels_set_speed(1,-1)
} else if(buffer[0]=='I') {
     differential_wheels_set_speed(-1,1)
}
else if(buffer[0]=='S') {
     differential_wheels_set_speed(0,0)
} else if (buffer[0]=='P') {


  matrix = gps_get_matrix(gps);
 sprintf(buffer,"%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d\r\n",
           gps_position_x(matrix),
	   gps_position_z(matrix),
	   gps_euler(matrix,euler),  
	    distance_sensor_get_value(ds[0]),
	    distance_sensor_get_value(ds[1]),
	    distance_sensor_get_value(ds[2]),
	    distance_sensor_get_value(ds[3]),
	    distance_sensor_get_value(ds[4]),
	    distance_sensor_get_value(ds[5]),
	    distance_sensor_get_value(ds[6]),
	    distance_sensor_get_value(ds[7]),
	    distance_sensor_get_value(ds[8]),
	    distance_sensor_get_value(ds[9]),
	    distance_sensor_get_value(ds[10]),
	    distance_sensor_get_value(ds[11]),
	    distance_sensor_get_value(ds[12]),
	    distance_sensor_get_value(ds[13]),
	    distance_sensor_get_value(ds[14]),
	    distance_sensor_get_value(ds[15]));
    send(fd,buffer,strlen(buffer),0);
     
}

  
/*if (buffer[0]=='D') { 
    sscanf(buffer,"D,%hd,%hd",&left_speed,&right_speed);
    differential_wheels_set_speed(left_speed,right_speed);
    send(fd,"d\r\n",3,0);
  } else if (buffer[0]=='L') {
    sscanf(buffer,"L,%hd,%hd",&led_number,&led_action);
    if (led_number>=0 && led_number<=1 && led_action>=0 && led_action<=2) {
      if (led_action==2) { 
	if (led_value[led_number]==1) led_action=0; /* switch off */
	else led_action=1; /* switch on */
      }
      led_value[led_number]=led_action;
      printf("set led %d to %d\n",led_number,led_action);
      led_set(led[led_number],(unsigned char)led_action);
    }
    send(fd,"l\r\n",3,0);
  } else if (buffer[0]=='G') { /* set the position counter */
    sscanf(buffer,"G,%hd,%hd",&left_speed,&right_speed);
    differential_wheels_set_encoders(left_speed,right_speed);
    send(fd,"g\r\n",3,0);
  } else if (buffer[0]=='B') { /* return a pretend version string */
    sprintf(buffer, "b,0,0\r\n");
    send(fd,buffer,strlen(buffer),0);
  } else if (buffer[0]=='N') { /* read distance sensor values */
    sprintf(buffer,"n,%d,%d,%d,%d,%d,%d,%d,%d\r\n",
	    distance_sensor_get_value(distance[0]),
	    distance_sensor_get_value(distance[1]),
	    distance_sensor_get_value(distance[2]),
	    distance_sensor_get_value(distance[3]),
	    distance_sensor_get_value(distance[4]),
	    distance_sensor_get_value(distance[5]),
	    distance_sensor_get_value(distance[6]),
	    distance_sensor_get_value(distance[7]));
    send(fd,buffer,strlen(buffer),0);
  } else if (buffer[0]=='H') {
    /* return the position counters of the pair of wheels */        
    int leftWheelEncoding = differential_wheels_get_left_encoder();
    int rightWheelEncoding = differential_wheels_get_right_encoder();
    sprintf(buffer, "h,%d,%d\r\n",leftWheelEncoding,rightWheelEncoding);
    send(fd,buffer,strlen(buffer),0);
  } else if (buffer[0]=='O') { /* read the ambient light sensor */ 
    /* CHECK - according to the Khepera II manual, the initial character */
    /* returned in 'n' - which seems like a document bug. */
    sprintf(buffer,"o,%d,%d,%d,%d,%d,%d,%d,%d\r\n",
	    light_sensor_get_value(ambient[0]),
	    light_sensor_get_value(ambient[1]),
	    light_sensor_get_value(ambient[2]),
	    light_sensor_get_value(ambient[3]),
	    light_sensor_get_value(ambient[4]),
	    light_sensor_get_value(ambient[5]),
	    light_sensor_get_value(ambient[6]),
	    light_sensor_get_value(ambient[7]));
    send(fd,buffer,strlen(buffer),0);                             
  } else if (strncmp(buffer,"exit",4)==0) {
    printf("connection closed\n");
#ifdef WIN32
    closesocket(fd);
    ret = WSACleanup();
#else
    ret = close(fd);
#endif
    if (ret!=0) fprintf(stderr,"Cannot close socket\n");
    fd=0;
  } else send(fd,"\n",1,0);


  return TIMESTEP;
}

/*----------------------------------------------------------------------*/
int main() {
  /* printf("*** Edinburgh University, IVR Course - TCPIP interface ***\n"); */
  robot_live(reset);
  robot_run(run);
  return 0;
}
